﻿using System;
// using System.Collections.Generic;
using UnityEngine;
using MonoMod.RuntimeDetour;
using Dungeonator;
using System.Reflection;

// Main Glitch Mod Module
namespace ChaosGlitchMod {

    public class GlitchModule : ETGModule {
        
        // Material static for GlitchShader materials used in GlitchMod
        private static Material Material;

        // Used by enterRoomHookGlitch hook
        private static RoomHandler room;

        // Hook statics
        private static Hook aihookGlitch;
        private static Hook hammerhookGlitch;
        private static Hook enterRoomHookGlitch;

        public override void Init() { }

        public override void Start() {
            // Modified version of Anywhere mod
            ETGModMainBehaviour.Instance.gameObject.AddComponent<DungeonFlowModule>();
            // Code for filling pots with enemies and upping Wall Mimic Spawns/other trolly things.
            ETGModMainBehaviour.Instance.gameObject.AddComponent<TrollModule>();

            ETGModConsole.Commands.AddGroup("glitch", delegate (string[] e)
            {
                ETGModConsole.Log("[Glitch Mode]  Valid sub-commands for glitchmode:\nenemies\nall\ntest\nreset\n\nRun reset command to turn off all glitch modes.", false);
            });


            ETGModConsole.Commands.GetGroup("glitch").AddUnit("enemies", delegate (string[] e)
            {
                bool aiFlag = aihookGlitch != null;
                bool hammerFlag = hammerhookGlitch != null;

                if (aiFlag) {
                    ETGModConsole.Log("Glitched Enemy mode already active! Use 'glitch reset' if you want to disable it!", false);
                } else {
                    aihookGlitch = new Hook(
                        typeof(AIActor).GetMethod("Awake"),
                        typeof(GlitchModule).GetMethod("AwakeHookGlitch")
                    );
                }
                if (hammerFlag) {
                    ETGModConsole.Log("Hammers already glitched!", false);
                } else {
                    hammerhookGlitch = new Hook(
                        typeof(ForgeHammerController).GetMethod("Activate"),
                        typeof(GlitchModule).GetMethod("HammerHookGlitch")
                    );
                    ETGModConsole.Log("Glitched Enemies Mode Activated...", false);
                }
            });

            ETGModConsole.Commands.GetGroup("glitch").AddUnit("all", delegate (string[] e)
            {
                bool aiFlag = aihookGlitch != null;
                bool hammerFlag = hammerhookGlitch != null;
                bool roomFlag = enterRoomHookGlitch != null;

                if (aiFlag) {
                    ETGModConsole.Log("Glitched Enemy mode already active!", false);
                } else {
                    aihookGlitch = new Hook(
                        typeof(AIActor).GetMethod("Awake"),
                        typeof(GlitchModule).GetMethod("AwakeHookGlitch")
                    );
                }
                if (hammerFlag) {
                    ETGModConsole.Log("Hammers already glitched!", false);
                } else {
                    hammerhookGlitch = new Hook(
                        typeof(ForgeHammerController).GetMethod("Activate"),
                        typeof(GlitchModule).GetMethod("HammerHookGlitch")
                    );
                }
                if (roomFlag) {
                    ETGModConsole.Log("Glitched Everything Mode already active!", false);
                } else {
                    enterRoomHookGlitch = new Hook(
                        typeof(RoomHandler).GetMethod("OnEntered", BindingFlags.NonPublic | BindingFlags.Instance),
                        typeof(GlitchModule).GetMethod("EnteredNewRoomHookForGlitch")
                    );
                    ETGModConsole.Log("Glitched Everything Mode Active...", false);
                }
            });

            ETGModConsole.Commands.GetGroup("glitch").AddUnit("test", delegate (string[] e) {
                foreach (BraveBehaviour s in UnityEngine.Object.FindObjectsOfType<BraveBehaviour>()) { BecomeGlitchedTest(s); }
                ETGModConsole.Log("One time glitch all. Enjoy the mess!", false);
            });

            ETGModConsole.Commands.GetGroup("glitch").AddUnit("reset", delegate (string[] e) {
                bool aiFlag = aihookGlitch != null;
                bool hammerFlag = hammerhookGlitch != null;
                bool roomFlag = enterRoomHookGlitch != null;

                if (aiFlag) { aihookGlitch.Dispose(); aihookGlitch = null; }
                if (hammerFlag) { hammerhookGlitch.Dispose(); hammerhookGlitch = null; }
                if (roomFlag) { enterRoomHookGlitch.Dispose(); enterRoomHookGlitch = null; }
                ETGModConsole.Log("Everything has returned to normal...", false);
            });

        }

        public override void Exit() { }

        // Glitch Everything Function (mostly written by Abeclancy with a dash of code from old Rainbow mod)
        public static void BecomeGlitched(BraveBehaviour s)
        {

            tk2dBaseSprite sprite = null;
            try
            {
                if (!(s.sprite is tk2dBaseSprite)) return;
                sprite = s.sprite;
            }
            catch { };

            if (s == null) return;
            if (sprite.usesOverrideMaterial) return;

            sprite.usesOverrideMaterial = true;
            Material = sprite.renderer.material;
            Material = new Material(ShaderCache.Acquire("Brave/Internal/Glitch"));
            Material.SetFloat("_GlitchInterval", 0.04f);
            Material.SetFloat("_DispProbability", 0.07f);
            Material.SetFloat("_DispIntensity", 0.05f);
            Material.SetFloat("_ColorProbability", 0.07f);
            Material.SetFloat("_ColorIntensity", 0.05f);
            // Material.SetFloat("_Cutoff", 1f);
            // Material.SetVector("_Color", new Vector4(0, 0, 0, 0));
            // Material.SetFloat("PixelSnap", 0f);
            // Material.SetFloat("_Perpendicular", 1f);
            // Material.SetFloat("_WrapDispCoords", 1f);
            // Material.SetFloat("_ColorGlitchOn", 1f);
            // Material.SetFloat("_DispGlitchOn", 1f);

            MeshRenderer component = sprite.GetComponent<MeshRenderer>();
            Material[] sharedMaterials = component.sharedMaterials;
            Array.Resize(ref sharedMaterials, sharedMaterials.Length + 1);
            Material material = UnityEngine.Object.Instantiate(Material);
            material.SetTexture("_MainTex", sharedMaterials[0].GetTexture("_MainTex"));
            sharedMaterials[sharedMaterials.Length - 1] = material;
            component.sharedMaterials = sharedMaterials;
        }

        public static void BecomeGlitchedTest(BraveBehaviour s)
        {

            tk2dBaseSprite sprite = null;
            try
            {
                if (!(s.sprite is tk2dBaseSprite)) return;
                sprite = s.sprite;
            }
            catch { };

            if (s == null) return;

            sprite.usesOverrideMaterial = true;
            Material = sprite.renderer.material;
            Material = new Material(ShaderCache.Acquire("Brave/Internal/Glitch"));
            Material.SetFloat("_GlitchInterval", 0.04f);
            Material.SetFloat("_DispProbability", 0.07f);
            Material.SetFloat("_DispIntensity", 0.05f);
            Material.SetFloat("_ColorProbability", 0.07f);
            Material.SetFloat("_ColorIntensity", 0.05f);

            MeshRenderer component = sprite.GetComponent<MeshRenderer>();
            Material[] sharedMaterials = component.sharedMaterials;
            Array.Resize(ref sharedMaterials, sharedMaterials.Length + 1);
            Material material = UnityEngine.Object.Instantiate(Material);
            material.SetTexture("_MainTex", sharedMaterials[0].GetTexture("_MainTex"));
            sharedMaterials[sharedMaterials.Length - 1] = material;
            component.sharedMaterials = sharedMaterials;
        }

        // Hook method for AIActor (enemies). Made with help from KyleTheScientist
        public static void AwakeHookGlitch(Action<AIActor> orig, AIActor self) {
            orig(self);


            Material = new Material(ShaderCache.Acquire("Brave/Internal/Glitch"));
            Material.SetFloat("_GlitchInterval", 0.04f);
            Material.SetFloat("_DispProbability", 0.12f);
            Material.SetFloat("_DispIntensity", 0.2f);
            Material.SetFloat("_ColorProbability", 0.1f);
            Material.SetFloat("_ColorIntensity", 0.15f);

            /*
            Material.SetFloat("_GlitchInterval", 0.1f);
            Material.SetFloat("_DispProbability", 0.4f);
            Material.SetFloat("_DispIntensity", 0.01f);
            Material.SetFloat("_ColorProbability", 0.4f);
            Material.SetFloat("_ColorIntensity", 0.04f);
            */

            MeshRenderer component = self.GetComponent<MeshRenderer>();
            Material[] sharedMaterials = component.sharedMaterials;
            Array.Resize(ref sharedMaterials, sharedMaterials.Length + 1);
            Material material = UnityEngine.Object.Instantiate(Material);
            material.SetTexture("_MainTex", sharedMaterials[0].GetTexture("_MainTex"));
            sharedMaterials[sharedMaterials.Length - 1] = material;
            component.sharedMaterials = sharedMaterials;

            // UnityEngine.Random random = new UnityEngine.Random();

            self.BaseMovementSpeed *= 1.1f;
            self.MovementSpeed *= 1.1f;

            HealthHaver healthHaver = self.healthHaver;
            if (healthHaver != null) { healthHaver.SetHealthMaximum(healthHaver.GetMaxHealth() / 1.5f, null, false); }
        }

        // Hook HammerController to apply GlitchShader.
        public static void HammerHookGlitch(Action<ForgeHammerController> orig, ForgeHammerController self) {
            orig(self);

            Material = new Material(ShaderCache.Acquire("Brave/Internal/Glitch"));
            MeshRenderer component = self.GetComponent<MeshRenderer>();
            Material[] sharedMaterials = component.sharedMaterials;
            Array.Resize(ref sharedMaterials, sharedMaterials.Length + 1);
            Material material = UnityEngine.Object.Instantiate(Material);
            material.SetTexture("_MainTex", sharedMaterials[0].GetTexture("_MainTex"));
            sharedMaterials[sharedMaterials.Length - 1] = material;
            component.sharedMaterials = sharedMaterials;

        }

        // Working hook of RoomHandler thanks to KyleTheScientist
        // Added check for new combat. Mostly to keep it from running this everytime I enter/exit a room.
        // Limit effect being re-applied to new combat rooms only.
        public static void EnteredNewRoomHookForGlitch(Action<RoomHandler, PlayerController> orig, RoomHandler self, PlayerController p) {
            orig(self, p);
            room = self;

            if (room.HasActiveEnemies(RoomHandler.ActiveEnemyType.RoomClear)) {
                foreach (BraveBehaviour s in UnityEngine.Object.FindObjectsOfType<BraveBehaviour>()) { BecomeGlitched(s); }
            }
        }

    }
}

