﻿using System;
using UnityEngine;
using MonoMod.RuntimeDetour;
using Gungeon;
using Dungeonator;
using System.Reflection;
using System.Collections.Generic;
using System.Linq;

namespace ChaosGlitchMod
{
    
    // Selects Random strings
    public static class ArrayExtensions {
        // This is an extension method. RandomEnemy() will now exist on all arrays.
        public static ENEMYRANDOMIZER RandomEnemy<ENEMYRANDOMIZER>(this ENEMYRANDOMIZER[] array) { return array[UnityEngine.Random.Range(0, array.Length)]; }
    }

    public class PotBehaviour : MonoBehaviour { private void FixedUpdate() { } }

    public class TrollModule : MonoBehaviour {
        
        private static Hook potenemyhook;
        private static Hook wallmimichook;

        // Set Max number of Wall mimics per room when super wall mimic mode is active.
        // These default values will be used for secret floors and the tutorial floor.
        public static int MaxWallMimicPerRoom = 2;
        public static int MaxWallMimicsForFloor = 40;
        public static bool debugMimicFlag = false;

        public static string[] PotEnemyList = {
            "rat",
            "rat_candle",
            "pot_fairy",
            "chicken",
            "snake",
            "bullat",
            "grenat",
            "shotgat",
            "spirat",
            "bullet_kings_toadie_revenge",
            "grenade_kin",
            "poopulons_corn",
            "robots_past_critter_1",
            "robots_past_critter_2",
            "robots_past_critter_3",
            "bullet_kings_toadie",
            "old_kings_toadie",
            "mine_flayers_claymore",
            "mine_flayers_bell",
            "dragun_egg_slimeguy",
            "tiny_blobulord",
            "fusebot",
            "poisbulin",
            "mouser",
            "blobulin",
            "rubber_kin",
            "convicts_past_soldier",
            "tazie",
            "NULL",
            "NULL2",
            "NULL3"
        };

        public static string[] PotEnemyListNoFairies = {
            "rat",
            "rat_candle",
            "chicken",
            "snake",
            "bullat",
            "grenat",
            "shotgat",
            "spirat",
            "bullet_kings_toadie_revenge",
            "grenade_kin",
            "poopulons_corn",
            "robots_past_critter_1",
            "robots_past_critter_2",
            "robots_past_critter_3",
            "bullet_kings_toadie",
            "old_kings_toadie",
            "mine_flayers_claymore",
            "mine_flayers_bell",
            "dragun_egg_slimeguy",
            "tiny_blobulord",
            "fusebot",
            "poisbulin",
            "mouser",
            "blobulin",
            "rubber_kin",
            "convicts_past_soldier",
            "tazie",
            "NULL",
            "NULL2",
            "NULL3"
        };

        public static string[] CritterList = {
            "rat",
            "rat_candle",
            "chicken",
            "snake",
            "robots_past_critter_1",
            "robots_past_critter_2",
            "robots_past_critter_3",
            "NULL",
            "NULL2"
        };

        public static string[] PestList = {
            "rat",
            "rat_candle",
            "snake",
            "dragun_egg_slimeguy",
            "tiny_blobulord",
            "NULL"
        };

        public static string[] TombstoneList = {
            "tombstoner",
            "NULL",
            "NULL2",
            "NULL3",
            "NULL4"
        };

        private void Init()	{ }

        private void Start()
        {
            bool WallMimicBoost = false;

            // SayNoToPots Mod by KyleTheScientist. Modified to allow random pool of enemies to spawn.
            ETGModConsole.Commands.AddGroup("chaos", delegate (string[] e)
            {
                ETGModConsole.Log("[Chaos Mode]  The following options are available for Chaos Mode:\npots\nwalls\nwalls_extreme\nnormal\nextreme\ndebug\nreset\n\nTo turn off all modes, use 'chaos reset'\nNote that changes to wall mimic settings will take effect on next floor load.", false);
            });

            ETGModConsole.Commands.GetGroup("chaos").AddUnit("pots", delegate (string[] e)
            {
                bool potFlag = potenemyhook != null;
                if (potFlag) {
                    ETGModConsole.Log("The Pots have already been filled! Use 'chaos reset' to disable it!", false);
                } else {
                    potenemyhook = new Hook(
                        typeof(MinorBreakable).GetMethod("OnBreakAnimationComplete", BindingFlags.Instance | BindingFlags.NonPublic),
                        typeof(TrollModule).GetMethod("SpawnAnnoyingEnemy")
                    );
                    ETGModConsole.Log("The Pots are filled...", false);
                }
            });

            ETGModConsole.Commands.GetGroup("chaos").AddUnit("walls", delegate (string[] e)
            {
                if (WallMimicBoost)
                {
                    WallMimicBoost = true;
                    ForceMimics();
                } else {
                    ETGModConsole.Log("The Walls are already untrusted!", false);
                }
                ETGModConsole.Log("The Walls can no longer be trusted...", false);
            });

            ETGModConsole.Commands.GetGroup("chaos").AddUnit("walls_extreme", delegate (string[] e)
            {
                bool superWallFlag = wallmimichook != null;
                if (!WallMimicBoost) { WallMimicBoost = true; ForceMimics(); }
                if (superWallFlag)
                {
                    ETGModConsole.Log("The Walls are already closing in! Use 'chaos reset' to disable it!", false);
                } else {
                    wallmimichook = new Hook(
                           typeof(Dungeon).GetMethod("PlaceWallMimics", BindingFlags.Public | BindingFlags.Instance),
                           typeof(TrollModule).GetMethod("PlaceWallMimicsHook", BindingFlags.Public | BindingFlags.Static)
                    );
                    ETGModConsole.Log("The Walls are closing in...", false);
                }
            });

            ETGModConsole.Commands.GetGroup("chaos").AddUnit("normal", delegate (string[] e)
            {
                bool potFlag = potenemyhook != null;
                if (potFlag) {
                    ETGModConsole.Log("The Pots have already been filled...", false);
                } else {
                    potenemyhook = new Hook(
                        typeof(MinorBreakable).GetMethod("OnBreakAnimationComplete", BindingFlags.Instance | BindingFlags.NonPublic),
                        typeof(TrollModule).GetMethod("SpawnAnnoyingEnemy")
                    );
                }
                if(!WallMimicBoost) { WallMimicBoost = true; ForceMimics(); }
                ETGModConsole.Log("Prepare to have a bad time! :D", false);
            });

            ETGModConsole.Commands.GetGroup("chaos").AddUnit("extreme", delegate (string[] e)
            {
                bool potFlag = potenemyhook != null;
                bool superWallFlag = wallmimichook != null;
                if (potFlag)
                {
                    ETGModConsole.Log("The Pots have already been filled...", false);
                }
                else
                {
                    potenemyhook = new Hook(
                        typeof(MinorBreakable).GetMethod("OnBreakAnimationComplete", BindingFlags.Instance | BindingFlags.NonPublic),
                        typeof(TrollModule).GetMethod("SpawnAnnoyingEnemy")
                    );
                }
                if (!WallMimicBoost)
                {
                    WallMimicBoost = true;
                    ForceMimics();
                } else {
                    ETGModConsole.Log("The Walls are already untrusted!", false);
                }
                if (superWallFlag)
                {
                    ETGModConsole.Log("The Walls are already closing in!", false);
                } else {
                    wallmimichook = new Hook(
                        typeof(Dungeon).GetMethod("PlaceWallMimics", BindingFlags.Public | BindingFlags.Instance),
                        typeof(TrollModule).GetMethod("PlaceWallMimicsHook", BindingFlags.Public | BindingFlags.Static)
                    );
                }
                ETGModConsole.Log("Prepare to have an extremely bad time! :D", false);
            });

            ETGModConsole.Commands.GetGroup("chaos").AddUnit("debug", delegate (string[] e)
            {

                if (!debugMimicFlag) {
                    debugMimicFlag = true;
                    ETGModConsole.Log("Debug Mode On...", false);
                } else {
                    if (debugMimicFlag) {
                        debugMimicFlag = false;
                        ETGModConsole.Log("Debug Mode Off...", false);
                    }
                }
            }); 

            ETGModConsole.Commands.GetGroup("chaos").AddUnit("reset", delegate (string[] e)
            {
                bool potFlag = potenemyhook != null;
                bool superWallFlag = wallmimichook != null;
                if (potFlag)
                {
                    potenemyhook.Dispose();
                    potenemyhook = null;
                }
                if (superWallFlag)
                {
                    wallmimichook.Dispose();
                    wallmimichook = null;
                }
                if (WallMimicBoost) { WallMimicBoost = false; ForceMimicsBack(); }
                ETGModConsole.Log("Everything has returned to normal...", false);
            });

        }

        private void Update() { }

        private void Exit() { }

        // SayNoToPots Mod - Now with ability to spawn more then one different type of enemy randomly! :D
        public static void SpawnAnnoyingEnemy(Action<MinorBreakable> orig, MinorBreakable self)
        {
            orig(self);

            PlayerController primaryPlayer = GameManager.Instance.PrimaryPlayer;
            
            // Special Enemy Pools
            bool flagTombstone = !self.name.ToLower().Contains("tombstone");
            bool flagBush = !self.name.ToLower().Contains("bush");

            // Enemy Pool with no Fairies
            bool flagBarrel = !self.name.ToLower().Contains("barrel");
            bool flagCrate = !self.name.ToLower().Contains("crate");
            bool flagSkull = !self.name.ToLower().Contains("skull");
            bool flagUrn = !self.name.ToLower().Contains("urn");
            bool flagWine = !self.name.ToLower().Contains("wine_stand");
            bool flagArmor = !self.name.ToLower().Contains("armor");
            bool flagBottle = !self.name.ToLower().Contains("bottle");
            bool flagIce = !self.name.ToLower().Contains("ice");
            // Full Enemy Pool
            bool flagPot = !self.name.ToLower().Contains("pot");

            // AIActor FullAIActorList = Game.Enemies[PotEnemyList.RandomEnemy()];
            // ETGModConsole.Log(self.name, false); RandomCritter

            if (primaryPlayer.CurrentRoom == self.sprite.WorldCenter.GetAbsoluteRoom())
            {
                if (primaryPlayer.IsInCombat)
                {
                    PotFairyEngageDoer.InstantSpawn = true;
                    if (!flagSkull | !flagUrn | !flagWine | !flagArmor)
                    {
                        AIActor.Spawn(Game.Enemies[PotEnemyListNoFairies.RandomEnemy()], self.sprite.WorldCenter, primaryPlayer.CurrentRoom, true, 0, true);
                    }
                    if (!flagPot)
                    {
                        AIActor.Spawn(Game.Enemies[PotEnemyList.RandomEnemy()], self.sprite.WorldCenter, primaryPlayer.CurrentRoom, true, 0, true);
                    }
                    if (!flagTombstone)
                    {
                        AIActor.Spawn(Game.Enemies[TombstoneList.RandomEnemy()], self.sprite.WorldCenter, primaryPlayer.CurrentRoom, true, 0, true);
                    }
                    if (!flagBush | !flagBottle)
                    {
                        AIActor.Spawn(Game.Enemies[CritterList.RandomEnemy()], self.sprite.WorldCenter, primaryPlayer.CurrentRoom, true, 0, true);
                    }
                    if (!flagCrate | !flagBarrel | !flagIce)
                    {
                        AIActor.Spawn(Game.Enemies[PestList.RandomEnemy()], self.sprite.WorldCenter, primaryPlayer.CurrentRoom, true, 0, true);
                    }
                } else { PotFairyEngageDoer.InstantSpawn = false; }
            }
        }

        // Wall Mimic Stuff courtasy of KyleTheScientist
        public void ForceMimics()
        {
            try
            {
                var injData = MetaInjectionData.WallMimicsAssignedToFloors;
                var values = (GlobalDungeonData.ValidTilesets[])Enum.GetValues(typeof(GlobalDungeonData.ValidTilesets));
                for (int i = 0; i < Enum.GetValues(typeof(GlobalDungeonData.ValidTilesets)).Length; i++)
                {
                    if (!injData.ContainsKey(values[i]))
                        injData.Add(values[i], 100);
                    else
                        injData[values[i]] = 100;
                }
            }
            catch (Exception e)
            {
                ETGModConsole.Log(e.Message);
            }

            BonusEnemySpawns bonusEnemySpawns = GameManager.Instance.RewardManager.WallMimicChances;
            bonusEnemySpawns.CastleChance = 1;
            bonusEnemySpawns.SewerChance = 1;
            bonusEnemySpawns.GungeonChance = 1;
            bonusEnemySpawns.CathedralChance = 1;
            bonusEnemySpawns.MinegeonChance = 1;
            bonusEnemySpawns.CatacombgeonChance = 1;
            bonusEnemySpawns.ForgegeonChance = 1;
            bonusEnemySpawns.BulletHellChance = 1;
        }

        public void ForceMimicsBack()
        {
            try
            {
                var injData = MetaInjectionData.WallMimicsAssignedToFloors;
                var values = (GlobalDungeonData.ValidTilesets[])Enum.GetValues(typeof(GlobalDungeonData.ValidTilesets));
                for (int i = 0; i < Enum.GetValues(typeof(GlobalDungeonData.ValidTilesets)).Length; i++)
                {
                    if (!injData.ContainsKey(values[i]))
                        injData.Add(values[i], 1);
                    else
                        injData[values[i]] = 1;
                }
            }
            catch (Exception e) { ETGModConsole.Log(e.Message); }

            BonusEnemySpawns bonusEnemySpawns = GameManager.Instance.RewardManager.WallMimicChances;
            bonusEnemySpawns.CastleChance = 1;
            bonusEnemySpawns.SewerChance = 1;
            bonusEnemySpawns.GungeonChance = 1;
            bonusEnemySpawns.CathedralChance = 1;
            bonusEnemySpawns.MinegeonChance = 1;
            bonusEnemySpawns.CatacombgeonChance = 1;
            bonusEnemySpawns.ForgegeonChance = 1;
            bonusEnemySpawns.BulletHellChance = 1;

        }

        public static void SetMimicStats()
        {
            int currentFloor = GameManager.Instance.CurrentFloor;

            // Secret Floors and Tutorial
            if (currentFloor == -1)
            {
                MaxWallMimicPerRoom = 2;
                MaxWallMimicsForFloor = 30;
            }
            // Normal Floors with 1 = Keep, 2 = Gungeon Proper, and so on
            // Floor 1 guranteed to have 1 mimic per room.
            if (currentFloor == 1)
            {
                
                MaxWallMimicPerRoom = 1;
                MaxWallMimicsForFloor = 20;
            }

            // Floor 2 and onwards can have more then one mimic per room.
            // However not a gurantee that every room will have that count.
            if (currentFloor == 2)
            {
                MaxWallMimicPerRoom = 2;
                MaxWallMimicsForFloor = 20;
            }

            if (currentFloor == 3)
            {
                MaxWallMimicPerRoom = 2;
                MaxWallMimicsForFloor = 25;
            }

            if (currentFloor == 4)
            {
                MaxWallMimicPerRoom = 3;
                MaxWallMimicsForFloor = 40;
            }

            if (currentFloor == 5)
            {
                MaxWallMimicPerRoom = 3;
                MaxWallMimicsForFloor = 50;
            }

            if (currentFloor >= 6)
            {
                MaxWallMimicPerRoom = 4;
                MaxWallMimicsForFloor = 60;
            }

            if (debugMimicFlag)
            {
                ETGModConsole.Log("[DEBUG] Current Floor: " + GameManager.Instance.CurrentFloor);
                ETGModConsole.Log("[DEBUG] Mimics Per Room: " + MaxWallMimicPerRoom);
                ETGModConsole.Log("[DEBUG] Mimics For Floor: " + MaxWallMimicsForFloor);
            }
            return;
        }

        public static void PlaceWallMimicsHook(Action<Dungeon, RoomHandler> orig, Dungeon self, RoomHandler debugRoom)
        {

            // Wall Mimics in the rat maze tile set doesn't work very well. Prevent them from spawning there now. Instead, have wall mimics on the tutorial level as a consolation price! :P
            // if (GameManager.Instance.CurrentLevelOverrideState != GameManager.LevelOverrideState.NONE && GameManager.Instance.CurrentLevelOverrideState != GameManager.LevelOverrideState.RESOURCEFUL_RAT) { return; }
            if (GameManager.Instance.CurrentLevelOverrideState != GameManager.LevelOverrideState.NONE && GameManager.Instance.CurrentLevelOverrideState != GameManager.LevelOverrideState.TUTORIAL) { return; }
            int numWallMimicsForFloor = MetaInjectionData.GetNumWallMimicsForFloor(self.tileIndices.tilesetId);
			if (numWallMimicsForFloor <= 0) { return; }

            List<int> list = Enumerable.Range(0, self.data.rooms.Count).ToList();
			list = list.Shuffle();

            // Set Max Wall Mimic values based on each floor. Secret floors and Tutorial are always -1 and will keep default values.
            SetMimicStats();

            if (debugRoom != null) { list = new List<int>(new int[] { self.data.rooms.IndexOf(debugRoom) }); }

            List<Tuple<IntVector2, DungeonData.Direction>> list2 = new List<Tuple<IntVector2, DungeonData.Direction>>();
            int num = 0;
			List<AIActor> list3 = new List<AIActor>();
			for (int i = 0; i < list.Count; i++)
			{
				RoomHandler roomHandler = self.data.rooms[list[i]];
                if (/*!roomHandler.IsShop &&*/ !roomHandler.IsMaintenanceRoom())
                {
                    if (!roomHandler.area.IsProceduralRoom || roomHandler.area.proceduralCells == null)
                    {
                        if (roomHandler.area.PrototypeRoomCategory != PrototypeDungeonRoom.RoomCategory.BOSS || !BraveUtility.RandomBool())
                        {
                            if (!roomHandler.GetRoomName().StartsWith("DraGunRoom"))
							{
								if (roomHandler.connectedRooms != null)
								{
									for (int j = 0; j < roomHandler.connectedRooms.Count; j++)
									{
										if (roomHandler.connectedRooms[j] == null || roomHandler.connectedRooms[j].area.PrototypeRoomCategory == PrototypeDungeonRoom.RoomCategory.BOSS) { }
									}
								}
								if (debugRoom == null)
								{
									bool flag = false;
									roomHandler.GetActiveEnemies(RoomHandler.ActiveEnemyType.All, ref list3);
									for (int k = 0; k < list3.Count; k++)
									{
										AIActor aiactor = list3[k];
										if (aiactor && aiactor.EnemyGuid == GameManager.Instance.RewardManager.WallMimicChances.EnemyGuid)
										{
											flag = true;
											break;
										}
									}
									if (flag) { goto IL_859; }
								}
								list2.Clear();
                                for (int l = -1; l <= roomHandler.area.dimensions.x; l++)
                                {
                                    for (int m = -1; m <= roomHandler.area.dimensions.y; m++)
                                    {
                                        int num2 = roomHandler.area.basePosition.x + l;
                                        int num3 = roomHandler.area.basePosition.y + m;
                                        if (self.data.isWall(num2, num3))
                                        {
                                            int n = 0;
                                            if (self.data.isWall(num2 - 1, num3) && self.data.isWall(num2 - 1, num3 - 1) && self.data.isWall(num2 - 1, num3 - 2) && self.data.isWall(num2, num3) && self.data.isWall(num2, num3 - 1) && self.data.isWall(num2, num3 - 2) && self.data.isPlainEmptyCell(num2, num3 + 1) && self.data.isWall(num2 + 1, num3) && self.data.isWall(num2 + 1, num3 - 1) && self.data.isWall(num2 + 1, num3 - 2) && self.data.isPlainEmptyCell(num2 + 1, num3 + 1) && self.data.isWall(num2 + 2, num3) && self.data.isWall(num2 + 2, num3 - 1) && self.data.isWall(num2 + 2, num3 - 2))
                                            {
                                                if (num2 % 3 == 0) list2.Add(Tuple.Create(new IntVector2(num2, num3), DungeonData.Direction.NORTH));
                                                if (num2 % 3 == 0) n++;
                                            }
                                            else if (self.data.isWall(num2 - 1, num3) && self.data.isWall(num2 - 1, num3 + 1) && self.data.isWall(num2 - 1, num3 + 2) && self.data.isWall(num2, num3) && self.data.isWall(num2, num3 + 1) && self.data.isWall(num2, num3 + 2) && self.data.isPlainEmptyCell(num2, num3 - 1) && self.data.isWall(num2 + 1, num3) && self.data.isWall(num2 + 1, num3 + 1) && self.data.isWall(num2 + 1, num3 + 2) && self.data.isPlainEmptyCell(num2 + 1, num3 - 1) && self.data.isWall(num2 + 2, num3) && self.data.isWall(num2 + 2, num3 + 1) && self.data.isWall(num2 + 2, num3 + 2))
                                            {
                                                if (num2 % 3 == 0) list2.Add(Tuple.Create(new IntVector2(num2, num3), DungeonData.Direction.SOUTH));
                                                if (num2 % 3 == 0) n++;
                                            }
                                            else if (self.data.isWall(num2, num3 + 2) && self.data.isWall(num2, num3 + 1) && self.data.isWall(num2, num3 - 1) && self.data.isWall(num2, num3 - 2) && self.data.isWall(num2 - 1, num3) && self.data.isPlainEmptyCell(num2 + 1, num3) && self.data.isPlainEmptyCell(num2 + 1, num3 - 1))
                                            {
                                                if (num3 % 3 == 0) list2.Add(Tuple.Create(new IntVector2(num2, num3), DungeonData.Direction.EAST));
                                                if (num3 % 3 == 0) n++;
                                            }
                                            else if (self.data.isWall(num2, num3 + 2) && self.data.isWall(num2, num3 + 1) && self.data.isWall(num2, num3 - 1) && self.data.isWall(num2, num3 - 2) && self.data.isWall(num2 + 1, num3) && self.data.isPlainEmptyCell(num2 - 1, num3) && self.data.isPlainEmptyCell(num2 - 1, num3 - 1))
                                            {
                                                if (num3 % 3 == 0) list2.Add(Tuple.Create(new IntVector2(num2 - 1, num3), DungeonData.Direction.WEST));
                                                if (num3 % 3 == 0) n++;
                                            }
                                            if (n > 0)
                                            {
                                                bool flag2 = true;
                                                int num4 = -5;
                                                while (num4 <= 5 && flag2)
                                                {
                                                    int num5 = -5;
                                                    while (num5 <= 5 && flag2)
                                                    {
                                                        int x = num2 + num4;
                                                        int y = num3 + num5;
                                                        if (self.data.CheckInBoundsAndValid(x, y))
                                                        {
                                                            CellData cellData = self.data[x, y];
                                                            if (cellData != null)
                                                            {
                                                                if (cellData.type == CellType.PIT || cellData.diagonalWallType != DiagonalWallType.NONE)
                                                                {
                                                                    flag2 = false;
                                                                }
                                                            }
                                                        }
                                                        num5++;
                                                    }
                                                    num4++;
                                                }
                                                if (!flag2)
                                                {
                                                    while (n > 0)
                                                    {
                                                        list2.RemoveAt(list2.Count - 1);
                                                        n--;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                if (debugRoom == null)
								{
                                    // Testing if this adds more wall mimics
                                    // ETGModConsole.Log("Number of Wall Mimics assigned to Floor: " + list.Count);
                                    for (int a = 0; a < MaxWallMimicPerRoom; ++a)
                                    {
                                        if (list2.Count > 0) 
                                        {
                                            Tuple<IntVector2, DungeonData.Direction> tuple = BraveUtility.RandomElement(list2);
                                            list2.Remove(tuple);
                                            IntVector2 first = tuple.First;
                                            DungeonData.Direction second = tuple.Second;
                                            if (second != DungeonData.Direction.WEST)
                                            {
                                                roomHandler.RuntimeStampCellComplex(first.x, first.y, CellType.FLOOR, DiagonalWallType.NONE);
                                            }
                                            if (second != DungeonData.Direction.EAST)
                                            {
                                                roomHandler.RuntimeStampCellComplex(first.x + 1, first.y, CellType.FLOOR, DiagonalWallType.NONE);
                                            }
                                            AIActor orLoadByGuid = EnemyDatabase.GetOrLoadByGuid(GameManager.Instance.RewardManager.WallMimicChances.EnemyGuid);
                                            AIActor.Spawn(orLoadByGuid, first, roomHandler, true, AIActor.AwakenAnimationType.Default, true);
                                            num++;
                                            if (num >= MaxWallMimicsForFloor) { break; }
                                        }
									}
								}
							}
						}
					}
				}
				IL_859:;
			}
			if (num > 0)
			{
				PhysicsEngine.Instance.ClearAllCachedTiles();
                if (debugMimicFlag) { ETGModConsole.Log("[DEBUG] Number of Mimics succesfully placed: " + num);  }
            }
        }
    }
}

